########################################################################
#
# File Name:            Restore.py
#
# Documentation:        http://docs.4suite.org/4SS/Lib/CommandLine.py.html
#
"""
WWW: http://4suite.org        e-mail: support@4suite.org

Copyright (c) 2001 Fourthought Inc, USA.   All Rights Reserved.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""


__doc__ = """Deserialize every object in a directory into 4SS"""

import os,string
from Ft.Server.Server import FtServerServerException, FTSERVER_NAMESPACE, Error
from Ft.Server.Console import CommandUtil, Authenticate
from Ft.Server.Core import Constants
from distutils import dir_util
from Ft.Server.Core.Lib import ConfigFile
from Ft.Xml import Domlette

from Ft.Xml import XPath

from Ft.Rdf.Serializers import Dom
from Ft.Rdf.Drivers import Memory
from Ft.Rdf import Model

def Run(options,args):

    d = args['dir']
    if not os.path.exists(d):
        raise Exception("Restore directory %s does not exist" % d)

    #Do authentication our selve
    options['AUTHENTICATION'] = Authenticate.UserAuthenticate(options=options)

    commit = 0
    try:
        repo = CommandUtil.GetRepository(options)

        ignored = ['/',
                   os.path.split(repo.getSystemUserUri())[0],
                   repo.getSystemUserUri(),
                   repo.getSystemSuperUserUri(),
                   repo.getSystemDocDefUri(),
                   repo.getSystemGroupUri(),
                   "%s/BASE_XML" % repo.getSystemDocDefUri(),
                   "%s/BASE_XSLT" % repo.getSystemDocDefUri(),
                   "%s/BASE_SCHEMATRON" % repo.getSystemDocDefUri(),
                   ]
        try:
            aliases = []
            print "Restore System Object"
            _DeserializeSystemObject(repo,d)

            print "Restore System Container"
            _DeserializeSystemContainer(repo,d,d,aliases,ignored)

            print "Restore Content"
            _DeserializeContainer(repo,d,d,aliases,ignored)
            print "Restore Aliases"
            for a in aliases:
                print "Restore Alias from %s" % a
                try:
                    repo.createFromXmlUri(a,update=1)
                except FtServerServerException, e:
                    if e.code == Error.UNKNOWN_URI:
                        print "Not Restoring broken alias %s" % a
                    else:
                        raise
            print "Restore RDF"
            fileName = os.path.join(d,'ftss-system.rdf')
            reader = Domlette.DEFAULT_NONVALIDATING_READER()
            szr = Dom.Serializer()
            dom = reader.fromUri(fileName)
            db = Memory.GetDb('test')
            db.begin()
            model = Model.Model(db)
            szr.deserialize(model, dom, '')

            sysModel = repo.getModel()

            for stmt in model.complete(None,None,None):
                subRes = sysModel.complete(stmt.subject,stmt.predicate,stmt.object)
                if not subRes:
                    print "Adding Statement (%s,%s,%s)" % (stmt.subject,stmt.predicate,stmt.object)
                    sysModel.add(stmt)


            commit = 1
        finally:
            if commit:
                repo.txCommit()
            else:
                repo.txRollback()
    except FtServerServerException, e:
        raise
        print e
        return

def _DeserializeContainer(repo,base,d,aliases,ignored):
    files = os.listdir(d)
    files.sort()
    for file in files:
        if file in ['ftss-system.rdf','ftss-system-object.xml']:
            #This is the RDF file
            continue
        fullPath = os.path.join(d,file)
        if os.path.isfile(fullPath):
            #We have a file!!!
            reader = Domlette.DEFAULT_NONVALIDATING_READER()
            dom = reader.fromUri(fullPath)
            uri = dom.documentElement.getAttributeNS(FTSERVER_NAMESPACE,'uri')

            #We ignore some URIs, see if this is one of them
            if uri in ignored:
                pass
            else:
                #Deserialize it
                #Make sure it is not an alias
                if dom.documentElement.namespaceURI == FTSERVER_NAMESPACE and dom.documentElement.localName == 'Alias':
                    aliases.append(fullPath)
                else:
                    if repo.hasResource(uri):
                        print "Update from %s" % fullPath
                    else:
                        print "Create from %s" % fullPath

                    repo.createFromXmlDom(dom,update=1)

            #Build any sub containers
            con = XPath.Context.Context(dom, processorNss={'f': FTSERVER_NAMESPACE})
            for child in XPath.Evaluate("/f:Container/f:Child/@uri",context = con):
                if child.value not in ignored:
                    #Recurse
                    newPath = os.path.join(base,child.value[1:])
                    if os.path.isdir(newPath):
                        _DeserializeContainer(repo,base,newPath,aliases,ignored)

        else:
            pass



def _DeserializeSystemContainer(repo,base,d,aliases,ignored):

    #Docdefs, users, groups

    print "Deserialize docdefs"
    _DeserializeContainer(repo,base,os.path.join(d,repo.getSystemDocDefUri()[1:]),aliases,ignored)

    print "Deserialize users"
    _DeserializeContainer(repo,base,os.path.join(d,repo.getSystemUserUri()[1:]),aliases,ignored)

    print "Deserialize groups"
    _DeserializeContainer(repo,base,os.path.join(d,repo.getSystemGroupUri()[1:]),aliases,ignored)


def _DeserializeSystemObject(repo,dir):

    reader = Domlette.DEFAULT_NONVALIDATING_READER()

    fileName = os.path.join(dir,'ftss-system-object.xml')

    dom = reader.fromUri(fileName)

    con = XPath.Context.Context(dom.documentElement, processorNss={'f': FTSERVER_NAMESPACE})
    acls = XPath.Evaluate('f:Acl', context=con)

    so = repo.getSystemObject()

    setAcls = {}
    for aclE in acls:
        for attr in aclE.attributes.values():
            if attr.name in setAcls.keys():
                so.addAcl(attr.name,attr.value)
            else:
                so.setAcl(attr.name,attr.value)
                setAcls[attr.name] = 1


def Register():
    from Ft.Lib.CommandLine import Options, Command, Arguments

    cmd = Command.Command('restore',
                          """Restore a backup of 4SS""",
                          '/tmp/4SSBackup',
                          __doc__,
                          function=Run,
                          arguments = [Arguments.RequiredArgument('dir',
                                                                  "The directory to read the backup information from",
                                                                  str),
                                       ],
                          )
    return cmd
