from Tkinter import *
from Graphics import *
from ComCtrls import *
from StdCtrls import *

def _enum(*args):
    num = 0
    for name in args:
        globals()[name] = num
        num = num + 1
    return

# MessageBox icons
mbError = 'error'
mbInfo = 'info'
mbQuestion = 'question'
mbWarning = 'warning'
# MessageBox buttons
mbAbort = 'abort'
mbRetry = 'retry'
mbIgnore = 'ignore'
mbAbortRetryIgnore = 'abortretryignore'
mbOk = 'ok'
mbCancel = 'cancel'
mbOkCancel = 'okcancel'
mbRetryCancel = 'retrycancel'
mbYesNo = 'yesno'
mbYes = 'yes'
mbNo = 'no'
mbYesNoCancel = 'yesnocancel'

# enum TModalResult {mrOk, mrCancel, mrAbort, mrRetry, mrIgnore, mrYes, mrNo };
mrNone = ''
mrOk = 'ok'
mrCancel = 'cancel'
mrAbort = 'abort'
mrRetry = 'retry'
mrIgnore = 'ignore'
mrYes = 'yes'
mrNo = 'no'

# enum TAlignment { taLeftJustify, taRightJustify, taCenter };
taLeftJustify=NW
taRightJustify=NE
taCenter=CENTER

# enum TBorderIcon
_enum('biSystemMenu', 'biMinimize', 'biMaximize', 'biHelp')

# enum TBorderStyle
_enum('bsNone',
      'bsSingle',
      'bsSizeable',
      'bsDialog',
      'bsToolWindow',
      'bsSizeToolWin')

# enum TEdgeBorder { ebLeft, ebTop, ebRight, ebBottom };
_enum('ebLeft',
      'ebTop',
      'ebRight',
      'ebBottom')

# enum TFormBorderStyle { bsNone, bsSingle, bsSizeable, bsDialog, bsToolWindow };
_enum('bsNone',
      'bsSingle',
      'bsSizeable',
      'bsDialog',
      'bsToolWindow')

# TWindowState
_enum('wsNormal',
      'wsMinimized',
      'wsMaximized')

# TFormStyle
_enum('fsNormal',
      'fsMDIChild',
      'fsMDIForm',
      'fsStayOnTop')

# TPosition
_enum('poDesigned',
      'poDefault',
      'poDefaultPosOnly',
      'poDefaultSizeOnly',
      'poScreenCenter',
      'poDesktopCenter')


# enum TScrollStyle { ssNone, ssHorizontal, ssVertical, ssBoth };
_enum('ssNone',
      'ssHorizontal',
      'ssVertical',
      'ssBoth',
      )


class TApplication(Tk):
    name = ""
    
    def __init__(self):
        title = self.name or self.__class__.__name__
        Tk.__init__(self)
        self.wm_withdraw()
        self.forms = []
        self.mainForm = None
       
    def run(self):
        if not self.mainForm:
            self.mainForm = self.forms[0]
        self.mainForm.show()
        self.mainloop()
        self.mainForm.wm_deiconify()
        self.terminate()
        return

    def terminate(self):
        # Time to clean up
        while self.forms:
            # Allow our forms to cleanup if they need to
            self.forms[-1].doDestroy()
            del self.forms[-1]

        # Don't know why we need this.
        # Without it, terminal focus goes into never-never land
        self.mainForm = None
        self.destroy()
        return

    def createForm(self, formClass):
        form = formClass(self)
        self.forms.append(form)
        form.onCreate and form.onCreate(form)
        return form

    def messageBox(self, title, message, icon=mbInfo, buttons=mbOk, default=mbOk):
        return self.tk.call('tk_messageBox',
                            '-icon', icon,
                            '-parent', self._w,
                            '-title', title,
                            '-message', message,
                            '-type', buttons,
                            '-default', default,
                            )

class TForm(Toplevel):
    onCanClose = None
    onClose = None
    onCreate = None
    onDestroy = None
    onShow = None
    onHide = None

    def __init__(self, owner, **options):
        Toplevel.__init__(self)
        self.owner = owner
        self.isModal = 0
        self.position = options.get('position', poDefault)
        self.top = options.get('top', 0)
        self.left = options.get('left', 0)
        self.width = options.get('clientWidth', 400)
        self.height = options.get('clientHeight', 300)
        self.clientArea = Frame(self)
        self.borderStyle = options.get('borderStyle', bsSizeable)
        if options.has_key('font'):
            self.font = options['font']
        else:
            self.font = TFont()
        if self.position == poScreenCenter:
            self.left = (self.winfo_screenwidth() - self.width) / 2
            self.top = (self.winfo_screenheight() - self.height) / 2
        self.geometry("%dx%d%+d%+d" % (
            self.width, self.height, self.left, self.top))
        self.withdraw()
        self.protocol("WM_DELETE_WINDOW", self.close)
        # For result when used with showModal()
        self._mrvar = StringVar()
        self.caption = options.get('caption', self.__class__.__name__)

        # Set our resizability
        if self.borderStyle == bsSizeable:
            self.tk.call('wm', 'resizable', self._w, '1', '1')
        else:
            self.tk.call('wm', 'resizable', self._w, '0', '0')

    def __setattr__(self, name, value):
        if name == 'caption':
            self.title(value)
        elif name == 'modalResult':
            self._mrvar.set(value)
        self.__dict__[name] = value
        return

    def doDestroy(self):
        self.onDestroy and self.onDestory(self)

    def show(self):
        self.focus()
        self.onShow and self.onShow(self)
        self.tk.call('wm', 'deiconify', self._w)
        
    def showModal(self):
        parent = self.focus_displayof()
        self._mrvar.set(mrNone)
        self.show()
        self.tk.call('grab', self._w)
        self.tk.call('tkwait', 'variable', self._mrvar)
        self.hide()
        parent and parent.show()
        return self._mrvar.get()
        
    def hide(self):
        self.onHide and self.onHide(self)
        self.tk.call('grab', 'release', self._w)
        self.tk.call('wm', 'withdraw', self._w)

    def close(self):
        if not self.onCanClose or self.onCanClose(self):
            self.onClose and self.onClose(self)
            self.quit()
        return

    redraw = Toplevel.update

class MenuBase:

    def enable(self,mode):
        if mode:
            self.owner.enable_child(self)
        else:
            self.owner.disable_child(self)
            
    def enable_child(self,menu):
        index = self._findChild(menu)
        if index != -1:
            if isinstance(self,TMainMenu): index = index + 1
            self.entryconfig(index,state='normal')
            return
        #Raise error??

    def disable_child(self,menu):
        index = self._findChild(menu)
        if index != -1:
            if isinstance(self,TMainMenu): index = index + 1
            self.entryconfig(index,state='disabled')
            return
        #Raise error??


    def _findChild(self,child):
        if child in self.winfo_children():
            return self.winfo_children().index(child)
        return -1

    
class TMainMenu(Menu,MenuBase):
    def __init__(self, owner):
        self.font = owner.font
        Menu.__init__(self, owner, font=self.font)
        owner.config(menu=self)
        self.label = 'Main Menu'



class TPopupMenu(Menu,MenuBase):
    def __init__(self, owner):
        self.font = owner.font
        Menu.__init__(self, owner, tearoff=0, font=self.font)

class TMenuItem(Menu,MenuBase):
    def __init__(self, owner, **options):
        self.owner = owner
        self.font = options.get('font', owner.font)
        Menu.__init__(self, owner, tearoff=0, font=self.font)
        label = options.get('label')
        (label, accel) = FindUnderline(label)
        self.label = label
        cascade = options.get('cascade')
        if options.has_key('cascade'): del options['cascade']

        if isinstance(owner, TMainMenu) or cascade:
            options['label'] = label
            options['underline'] = accel
            options['menu'] = self
            apply(owner.add_cascade, (), options)
        elif label == '-':
            owner.add_separator()
        else:
            options['label'] = label
            options['underline'] = accel
            apply(owner.add_command, (), options)



class TListView:
    # Always uses Report style
    def __init__(self, owner, **options):
        self.owner = owner
        self.items = TListItems(self)
        self.columns = options.get('columns', [])
        self.left = options.get('left', 8)
        self.top = options.get('top', 8)
        self.width = options.get('width', 250)
        self.height = options.get('height', 150)
        self.showColumnHeaders = options.get('showColumnHeaders', 1)
        self.columnClick = self.showColumnHeaders and options.get('columnClick', 1)
        self.popupMenu = options.get('popupMenu')
        self.onSelect = options.get('onSelect')
        self.onDblClick = options.get('onDblClick')
        self.onColumnClick = options.get('onColumnClick')
        self.scrollOption = options.get('scrollbar',ssNone)
        self.selected = None
        
        # Create the widgets
        self.frame = Frame(owner, relief=SUNKEN, bd=2)
        self.frame.place(x=self.left, y=self.top,
                         width=self.width, height=self.height)

        
        if self.scrollOption in [ssVertical, ssBoth]:
            self.vScrollbar = Scrollbar(self.frame,
                                        orient=VERTICAL,
                                        command=self.doVScroll)
            self.vScrollbar.pack(side=RIGHT, fill=Y)
            self.vScrollWidth = self.vScrollbar.winfo_reqwidth()
        else:
            self.vScrollbar = None
            self.vScrollWidth = 0

        for column in self.columns:
            column.bounds = Frame(self.frame,
                                  borderwidth=0,
                                  background=clWindow)
            if self.showColumnHeaders:
                column.header = Label(column.bounds,
                                      text=column.caption,
                                      anchor=column.alignment,
                                      relief=RAISED,
                                      borderwidth=2,
                                      font=owner.font)
                if self.columnClick:
                    column.header.bind('<ButtonPress-1>', self.doColumnPress)
                    column.header.bind('<B1-Motion>', self.doColumnMove)
                    column.header.bind('<ButtonRelease-1>', self.doColumnClick)
            column.listbox = Listbox(column.bounds,
                                     background=clWindow,
                                     borderwidth=0,
                                     yscrollcommand=self.doColumnScroll,
                                     exportselection=NO,
                                     font=owner.font)
            column.listbox.bind('<Button-1>', self.doClick)
            column.listbox.bind('<Double-Button-1>', self.doDblClick)
            column.listbox.bind('<Button-3>', self.doMouseDown)

            if self.scrollOption in [ssHorizontal, ssBoth]:
                column.hbar = Scrollbar(column.bounds,
                                        orient=HORIZONTAL,
                                        command=column.listbox.xview)
                column.listbox.config(xscrollcommand = column.hbar.set)
            else:
                column.hbar = None

        # Inner width is width - scrollbar width - border width (L & R)
        innerWidth = self.width - self.vScrollWidth - 4
        auto = 0
        for column in self.columns:
            if not column.autoSize:
                innerWidth = innerWidth - column.width
            else:
                auto = auto + 1
        # Probably want to add some smarts about uneven widths
        autoWidth = innerWidth / auto
        left = 0
        for column in self.columns:
            if column.autoSize:
                column.width = autoWidth
            column.bounds.place(x=left, relheight=1, width=column.width)
            if self.showColumnHeaders:
                column.header.pack(fill=X)
            column.listbox.pack(fill=BOTH,expand=1)
            if column.hbar:
                column.hbar.pack(fill=X, expand=1)
            left = left + column.width
        return
    
    def doResize(self):
        return

    def redraw(self):
        # Create a list of lists (one for each column)
        contents = map(lambda c: [END], self.columns)
        for item in self.items:
            contents[0].append(item.caption)
            for column in range(1,len(self.columns)):
                data = contents[column]
                if column <= len(item.subitems):
                    data.append(item.subitems[column-1])
                else:
                    data.append('')
            
        # Save the current position in the display
        (top, bottom) = self.columns[0].listbox.yview()

        # Empty the current contents
        for col in range(len(self.columns)):
            self.columns[col].listbox.delete(0, END)
            apply(self.columns[col].listbox.insert, contents[col])

        # Restore the display position
        self.doColumnScroll(top, bottom)

        if self.selected:
            self.doSelect(self.items.indexOf(self.selected))

        # Force an update of the lists
        self.frame.update()
        return

    def doVScroll(self, *args):
        # Called when vertical scrollbar is moved
        for column in self.columns:
            apply(column.listbox.yview, args)
        return

    def doColumnScroll(self, top, bottom):
        # Called when listbox contents are scrolled
        if self.vScrollbar:
            self.vScrollbar.set(top, bottom)
        for column in self.columns:
            column.listbox.yview('moveto', top)
        return        

    def doClick(self, event):
        # Primary mouse button
        index = int(event.widget.nearest(event.y))
        self.doSelect(index)
        return

    def doSelect(self, index):
        if index < 0 or index > len(self.items): return
        self.selected = self.items[index]
        for column in self.columns:
            column.listbox.selection_clear(0, len(self.items)-1)
            column.listbox.selection_set(index)
        self.onSelect and self.onSelect(self.owner,self.selected)
        return

    def doDblClick(self, event):
        # Primary mouse button
        index = int(event.widget.nearest(event.y))
        if index < 0 or index > len(self.items): return
        self.selected = self.items[index]
        self.onDblClick and self.onDblClick(self.owner,self.selected)
        return

    def doMouseDown(self, event):
        self.doClick(event)
        index = int(event.widget.nearest(event.y))
        if index < 0 or index > len(self.items): return
        if self.popupMenu:
            self.popupMenu.tk_popup(event.x_root, event.y_root)
        return

    # The following doColumn... methods simulate clickable button
    def doColumnPress(self, event):
        event.widget.config(relief=SUNKEN)
        return

    def doColumnMove(self, event):
        containing = event.widget.winfo_containing(event.x_root, event.y_root)
        if containing != event.widget:
            event.widget.config(relief=RAISED)
        else:
            event.widget.config(relief=SUNKEN)
        return

    def doColumnClick(self, event):
        if event.widget.cget('relief') == SUNKEN:
            # Shortcut determining if mouse is in the header
            event.widget.config(relief=RAISED)
            # Primary mouse button
            for column in self.columns:
                if column.header == event.widget:
                    # Add callback here
                    self.onColumnClick and self.onColumnClick(column)
                    return
        return

class TListColumn:
    def __init__(self, **options):
        self.alignment = options.get('alignment', taLeftJustify)
        self.autoSize = options.get('autoSize', 0)
        self.caption = options.get('caption', '')
        self.displayName = options.get('displayName', '')
        self.imageIndex = options.get('imageIndex', 0)
        self.maxWidth = options.get('maxWidth', 0)
        self.minWidth = options.get('minWidth', 0)
        self.width = options.get('width', 50)
        self.index = -1

# enum TToolButtonStyle { tbsButton, tbsCheck, tbsDropDown, tbsSeparator, tbsDivider };
_enum('tbsButton',
      'tbsCheck',
      'tbsDropDown',
      'tbsSeparator',
      'tbsDivider')

class TToolBar:
    def __init__(self, owner, **options):
        return

class TToolButton:
    def __init__(self, owner, **options):
        return

class TStatusBar:
    def __init__(self, owner, **options):
        self.owner = owner
        self.left = options.get('left', 8)
        self.top = options.get('top', 8)
        self.width = options.get('width', 250)
        self.height = options.get('height', 150)
        self.panels = options.get('panels', [])
        self.frame = Frame(owner)
        self.frame.place(x=self.left, y=self.top,
                         width=self.width, height=self.height)
        self.simpleText = Label(self.frame,
                                bd=1,
                                relief=SUNKEN,
                                )
        self.simpleText.place(y=1, height=self.height-1, relwidth=1)
