
from Ft.Server.Core import Constants, CreationParamsImp
import FtssTypes

def CreationParamsToCorba(cp):
    #Translate them
    if isinstance(cp,CreationParamsImp.CreationParams):

        fti = cp.fullTextIndex
        if cp.creationFlags == Constants.CreationFlags.CREATE_AS_XML:
            validationInfo = ValidationInfoToCorba(cp.validationInfo)
            klass = FtssTypes.XML_DOCUMENT
            validate = 0
            extensionModules = []
        elif cp.creationFlags == Constants.CreationFlags.CREATE_AS_XSLT:
            validationInfo = ValidationInfoToCorba(None)
            klass = FtssTypes.XSLT_DOCUMENT
            validate = cp.validate
            extensionModules = cp.extModules
        elif cp.creationFlags == Constants.CreationFlags.CREATE_AS_SCHEMATRON:
            validationInfo = ValidationInfoToCorba(None)
            klass = FtssTypes.SCHEMATRON_DOCUMENT
            validate = cp.validate
            extensionModules = []

        return FtssTypes.CreationParameters(fti,klass,validationInfo,validate,extensionModules)
    return cp
    

def ValidationInfoToCorba(vi):
    if vi is None or vi.validationType == Constants.ValidationType.NONE:
        return FtssTypes.Validator(FtssTypes.NONE,FtssTypes.DocumentReference(FtssTypes.EMPTY_REFERENCE,'','',''))
    elif vi.validationType == Constants.ValidationType.SCHEMATRON:
        dr = DocumentReferenceToCorba(vi.documentReference)
        return FtssTypes.Validator(FtssTypes.SCHEMATRON,dr)
    elif vi.validationType == Constants.ValidationType.DTD:
        dr = DocumentReferenceToCorba(vi.documentReference)
        return FtssTypes.Validator(FtssTypes.DTD,dr)

def DocumentReferenceToCorba(dr):
    if dr.referenceType == Constants.DocumentReferenceType.STRING:
        t = FtssTypes.DOCUMENT_STRING
        uri = ''
        baseUri = dr.baseUri
        source = dr.data
    elif dr.referenceType == Constants.DocumentReferenceType.EXTERNAL:
        t = FtssTypes.EXTERNAL
        uri = dr.uri
        baseUri = ''
        source = ''
    elif dr.referenceType == Constants.DocumentReferenceType.INTERNAL:
        t = FtssTypes.INTERNAL
        uri = dr.uri
        baseUri = ''
        source = ''
    return FtssTypes.DocumentReference(t,uri,source,baseUri)


def NsMapToCorba(nss):
    if type(nss) == type({}):
        newNss = []
        for n,value in nss.items():
            ns = FtssTypes.NamespaceMap(n,value)
            newNss.append(ns)
        nss = newNss
    return nss


def RdfMapsToCorba(maps):
    if maps and type(maps[0]) == type(()):
        newMaps = []
        for sub,pred,obj in maps:
            m = FtssTypes.RdfMap(sub,pred,obj)
            newMaps.append(m)
        maps = newMaps
    return maps
