import sys, os, socket, threading, time
import CORBA, PortableServer, CosNaming
from omniORB import CORBA
from Ft.Server.Server import FtServerServerException, Error
from Ft.Server.Protocols import LogUtil, Launcher, Controller

from Ft.Server.Protocols.CORBA.FtOmniorb import FtServerFactoryImp
from Ft.Server.Protocols.CORBA import FtServerCorbaException, CorbaError
from Ft.Server.Protocols.CORBA.FtOmniorb.Lib import CorbaUtil

DEBUG = 0

REAPER_DELAY = .75

class OmniorbServer(Controller.Controller):
    def __init__(self, args):
        Controller.Controller.__init__(self, 'OmniorbServer', args)

    def _initialize(self):
        # Called before startup
        self._reaperThread = threading.Thread(None,self._reaperFunc)
        self._runThread = 1
        self._reaperThread.start()
        self._factory = None

        self._reload()
        return

    def _run(self):
        self._orb.run()

    def _stop(self):
        self._orb.shutdown(1)
        self._runThread = 0
        return

    def _reload(self):

        host = self._properties['HostName'] or socket.gethostname()
        name_port = '%s:%s' % (host, self._properties['Port'])
        args = ['', '-ORBpoa_iiop_name_port', name_port]
        if DEBUG:
            args.extend(['-ORBtraceLevel', '10'])
        # FIXME - Should we specify UseHost option?
        self._orb = orb = CORBA.ORB_init(args, CORBA.ORB_ID)
        poa = orb.resolve_initial_references("RootPOA")
        poaManager = poa._get_the_POAManager()
        poaManager.activate()

        try:
            ns = orb.resolve_initial_references("NameService")
        except CORBA.NO_RESOURCES:
            self._runThread = 0
            raise FtServerCorbaException(CorbaError.NAME_SERVICE_NOT_RUNNING)

        # Create the object
        self._factory = imp = FtServerFactoryImp.FtServerFactoryImp(self._properties,self._configFile,self._coreId)
        object = imp._this()

        #Register this server with the Nameing Service
        #FIXME Get from config file
        nc = CosNaming.NameComponent('FtServerFactory', '')
        ns.rebind([nc], object)
        return

    def _reaperFunc(self):
        while self._runThread:
            time.sleep(REAPER_DELAY)
            if self._factory:
                self._factory._reap()
        

if __name__ == '__main__':
    controller = OmniorbServer(sys.argv[1:])
    controller.process_run()
    
