#include "domlette.h"

#define CHILDNODES_CHECK(doc) if (!PyDocument_Check((doc)) || !((PyDocumentObject *)(doc))->childNodes) {PyErr_SetString(g_errorObject, "Document in inconsistent state"); return NULL;}
#define DOCELEM_CHECK(doc) if (!PyDocument_Check((doc)) || !((PyDocumentObject *)(doc))->documentElement) {PyErr_SetString(g_errorObject, "Document in inconsistent state"); return NULL;}
#define DOCTYPE_CHECK(doc) if (!PyDocument_Check((doc)) || !((PyDocumentObject *)(doc))->docType) {PyErr_SetString(g_errorObject, "Document in inconsistent state"); return NULL;}
#define DOCUMENT_CHECK(doc) CHILDNODES_CHECK((doc)); DOCELEM_CHECK((doc)); DOCTYPE_CHECK((doc));

PyObject *g_documentNodeType; // added by Brendan Larson 9/12/2002 for Mac OS X 10.1.x compiler compatiblity
PyObject *g_documentNodeName; // added by Brendan Larson 9/12/2002 for Mac OS X 10.1.x compiler compatiblity


/*Internal Methods */

/* No need to intern the empty string, Python already does this */
#define UNICODE_EMPTY_STRING PyUnicode_FromUnicode(NULL, 0)

static PyObject *InternUnicode(PyDocumentObject *doc, PyObject *unistr)
{
#ifdef DISABLE_INTERN
  Py_INCREF(unistr);
  return unistr;
#else
  PyObject *v = unistr;

  if (unistr != Py_None) {
    PyObject *interned = doc->internedUnicode;
    if ((v = PyDict_GetItem(interned, unistr)) == NULL) {
      PyErr_Clear();
      if (PyDict_SetItem(interned, unistr, unistr)) {
	PyErr_SetString(PyExc_SystemError, "InternUnicode: intern failed");
	return NULL;
      }
      v = unistr;
    }
  }
    

  Py_INCREF(v);
  return v;
#endif
}

static PyObject *MakeQName(PyDocumentObject *doc, PyObject *prefix, PyObject *localName)
{
  PyObject *nodeName;

  if (PyObject_IsTrue(prefix)) {
    PyObject *u, *v;
    u = PyUnicode_DecodeASCII(":", 1, NULL);
    if (u == NULL) return NULL;
    
    v = PySequence_Concat(prefix, u);
    Py_DECREF(u);
    if (v == NULL) return NULL;

    u = PySequence_Concat(v, localName);
    Py_DECREF(v);
    if (u == NULL) return NULL;

    nodeName = InternUnicode(doc, u);
    Py_DECREF(u);
  } else {
    nodeName = InternUnicode(doc, localName);
  }
  return nodeName;
}


PyDocumentObject *Document_NEW(unsigned long *docIx, PyObject *refUri)
     /* Returns a new ref */
{
  register PyDocumentObject *doc;

  doc = PyObject_GC_New(PyDocumentObject, &PyDomletteDocument_Type);
  if (doc == NULL) {
    return NULL;
  }

  Node_INIT((PyNodeObject*)doc, Py_None, &PyDomletteDocument_Type);

  doc->documentElement = Py_None;
  Py_XINCREF(Py_None);

  doc->docIndex = (*docIx)++;

  if (refUri == NULL) 
    refUri = UNICODE_EMPTY_STRING; /* new reference */
  else
    Py_INCREF(refUri);
  doc->refUri = refUri;

  doc->internedUnicode = PyDict_New();
  doc->childNodes = PyList_New(0);

  doc->docType = Py_None;
  Py_INCREF(Py_None);

  PyObject_GC_Track(doc);

  return doc;
}


PyElementObject *Document_CreateElementNS(PyDocumentObject *doc,
					  PyObject *namespaceURI,
					  PyObject *prefix,
					  PyObject *localName,
					  unsigned long *docIx)
{
  /* returns new ref */
  register PyElementObject *ele;
  PyObject *nodeName;

  DOCUMENT_CHECK(doc);

  NULL_PREFIX_CHECK(prefix, "Document_CreateElementNS");
  NULL_NAMESPACE_CHECK(namespaceURI, "Document_CreateElementNS");

  namespaceURI = InternUnicode(doc, namespaceURI);
  if (!namespaceURI) return NULL;

  prefix = InternUnicode(doc, prefix);
  if (!prefix) {
    Py_DECREF(namespaceURI);
    return NULL;
  }

  localName = InternUnicode(doc, localName);
  if (!localName) {
    Py_DECREF(prefix);
    Py_DECREF(namespaceURI);
    return NULL;
  }

  nodeName = MakeQName(doc, prefix, localName);
  if (!nodeName) {
    Py_DECREF(localName);
    Py_DECREF(prefix);
    Py_DECREF(namespaceURI);
    return NULL;
  }

  ele = PyObject_GC_New(PyElementObject, &PyDomletteElement_Type);
  if (ele == NULL) {
    Py_DECREF(nodeName);
    Py_DECREF(localName);
    Py_DECREF(prefix);
    Py_DECREF(namespaceURI);
    return NULL;
  }

  Node_INIT((PyNodeObject*)ele, (PyObject *)doc,&PyDomletteElement_Type);

  ele->attributes = PyDict_New();
  ele->childNodes = PyList_New(0);
  ele->docIndex = *docIx;
  /* Make space for namespace nodes index (between element and attr) */
  (*docIx)+=2;

  ele->namespaceURI = namespaceURI;
  ele->prefix = prefix;
  ele->localName = localName;
  ele->nodeName = nodeName;

  PyObject_GC_Track(ele);

  return ele;
}


PyAttrObject *Document_CreateAttributeNS(PyDocumentObject *doc,
					 PyObject *namespaceURI,
					 PyObject *localName,
					 PyObject *prefix,
                                         PyObject *value,
					 unsigned long *docIx)
{
  /* returns new ref */
  register PyAttrObject *attr;
  PyObject *nodeName;

  DOCUMENT_CHECK(doc);

  NULL_PREFIX_CHECK(prefix, "Document_CreateAttributeNS");
  NULL_NAMESPACE_CHECK(namespaceURI, "Document_CreateAttributeNS");

  namespaceURI = InternUnicode(doc, namespaceURI);
  if (!namespaceURI) return NULL;

  prefix = InternUnicode(doc, prefix);
  if (!prefix) {
    Py_DECREF(namespaceURI);
    return NULL;
  }

  localName = InternUnicode(doc, localName);
  if (!localName) {
    Py_DECREF(prefix);
    Py_DECREF(namespaceURI);
    return NULL;
  }

  if (value == NULL)
    value = UNICODE_EMPTY_STRING; /* new reference */
  else
    value = InternUnicode(doc, value); /* new reference */
  if (!value) {
    Py_DECREF(localName);
    Py_DECREF(prefix);
    Py_DECREF(namespaceURI);
    return NULL;
  }

  nodeName = MakeQName(doc, prefix, localName);
  if (!nodeName) {
    Py_DECREF(value);
    Py_DECREF(localName);
    Py_DECREF(prefix);
    Py_DECREF(namespaceURI);
    return NULL;
  }

  attr = PyObject_GC_New(PyAttrObject, &PyDomletteAttr_Type);
  if (attr == NULL) {
    Py_DECREF(nodeName);
    Py_DECREF(value);
    Py_DECREF(localName);
    Py_DECREF(prefix);
    Py_DECREF(namespaceURI);
    return NULL;
  }

  Node_INIT((PyNodeObject*)attr, (PyObject *)doc, &PyDomletteAttr_Type);

  attr->namespaceURI = namespaceURI;
  attr->prefix = prefix;
  attr->localName = localName;
  attr->nodeName = nodeName;
  attr->nodeValue = value;

  /*attr->docIndex = (*docIx)++;*/
  /* Attributes are unordered, so don't increment */
  attr->docIndex = *docIx;

  PyObject_GC_Track(attr);

  return attr;
}


PyTextObject *Document_CreateTextNode(PyDocumentObject *doc,
                                      PyObject *data,
                                      unsigned long *docIx)
{
  /*Returns new reference*/
  register PyTextObject *text;

  DOCUMENT_CHECK(doc);

  data = InternUnicode(doc, data);
  if (data == NULL) {
    return NULL;
  }

  text = PyObject_GC_New(PyTextObject, &PyDomletteText_Type);
  if (text == NULL) {
    Py_DECREF(data);
    return NULL;
  }

  Node_INIT((PyNodeObject*)text, (PyObject *)doc, &PyDomletteText_Type);

  text->nodeValue = data;
  text->docIndex = (*docIx)++;

  PyObject_GC_Track(text);

  return text;
}


PyCommentObject *Document_CreateComment(PyDocumentObject *doc,
                                        PyObject *data,
                                        unsigned long *docIx)
{
  /*Returns new reference*/
  register PyCommentObject *comment;

  DOCUMENT_CHECK(doc);

  data = InternUnicode(doc, data);
  if (data == NULL) {
    return NULL;
  }

  comment = PyObject_GC_New(PyCommentObject, &PyDomletteComment_Type);
  if (comment == NULL) {
    Py_DECREF(data);
    return NULL;
  }

  Node_INIT((PyNodeObject*)comment, (PyObject *)doc,&PyDomletteComment_Type);

  comment->nodeValue = data;
  comment->docIndex = (*docIx)++;

  PyObject_GC_Track(comment);
  return comment;
}


PyProcessingInstructionObject *Document_CreateProcessingInstruction(
                                    PyDocumentObject *doc,
                                    PyObject *target,
                                    PyObject *data,
				    unsigned long *docIx)
{
  /*Returns new reference*/
  register PyProcessingInstructionObject *pi;

  DOCUMENT_CHECK(doc);

  target = InternUnicode(doc, target);
  if (target == NULL) {
    return NULL;
  }

  data = InternUnicode(doc, data);
  if (data == NULL) {
    Py_DECREF(target);
    return NULL;
  }
  
  pi = PyObject_GC_New(PyProcessingInstructionObject, &PyDomletteProcessingInstruction_Type);
  if (pi == NULL) {
    Py_DECREF(target);
    Py_DECREF(data);
    return NULL;
  }

  Node_INIT((PyNodeObject*)pi, (PyObject *)doc,&PyDomletteProcessingInstruction_Type);

  pi->target = target;
  pi->nodeValue = data;
  pi->docIndex = (*docIx)++;

  PyObject_GC_Track(pi);

  return pi;
}


PyDocumentFragmentObject *Document_CreateDocumentFragment(PyDocumentObject *doc,
							  unsigned long *docIx)
{
  /*Returns new reference*/
  register PyDocumentFragmentObject *df;

  DOCUMENT_CHECK(doc);

  df = PyObject_GC_New(PyDocumentFragmentObject, &PyDomletteDocumentFragment_Type);
  if (df == NULL) {
    return NULL;
  }

  Node_INIT((PyNodeObject*)df, (PyObject *)doc,&PyDomletteDocumentFragment_Type);

  df->docIndex = (*docIx)++;
  df->childNodes = PyList_New(0);

  PyObject_GC_Track(df);

  return df;
}



/*External Methods*/
static char PyDocument_createElementNS__doc__[] =
"Create a new element";
PyObject *PyDocument_createElementNS(PyObject *self, PyObject * args) {

  PyObject *res;
  unsigned long index;
  PyObject *namespaceUri, *qName, *localName, *prefix;

  DOCUMENT_CHECK(self);

  if(!PyArg_ParseTuple(args, "OO:createElementNS", &namespaceUri, &qName))
    return NULL;

  if (namespaceUri != Py_None) {
    namespaceUri = PyUnicode_FromObject(namespaceUri);
    if (!namespaceUri ) {
      return NULL;
    }
  } else {
    Py_XINCREF(Py_None);
  }

  /*Don't intern this, we are plus one on qName*/
  qName = PyUnicode_FromObject(qName);
  if (!qName) {
    Py_DECREF(namespaceUri);
    return NULL;
  }

  /* Split the qname into localname and prefix*/
  /* After the split, we have an interned prefix (or None) and qname, each with a new ref */
  SplitQName(qName, &prefix, &localName);

  res = (PyObject *)Document_CreateElementNS((PyDocumentObject *)self,
                                             namespaceUri,
                                             prefix,
                                             localName,
                                             &index);
  Py_DECREF(namespaceUri);
  Py_DECREF(prefix);
  Py_DECREF(localName);
  /*Create Element creates its own nodeName */

  /*Remove our ref*/
  Py_DECREF(qName);
  return res;
}


static char PyDocument_createAttributeNS__doc__[] =
"Create a new attribute";
PyObject *PyDocument_createAttributeNS(PyObject * self, PyObject * args) {

  PyObject *res;
  unsigned long index;
  PyObject *namespaceUri, *qName, *localName, *prefix;

  DOCUMENT_CHECK(self);

  if(!PyArg_ParseTuple(args,"OO:createAttributeNS",&namespaceUri,&qName))
    return NULL;


  if (namespaceUri != Py_None) {
    namespaceUri = PyUnicode_FromObject(namespaceUri);
    if (!namespaceUri ) {
      return NULL;
    }
  } else {
    Py_XINCREF(Py_None);
  }

  qName = PyUnicode_FromObject(qName);
  if (!qName) {
    Py_DECREF(namespaceUri);
    return NULL;
  }

  /*Split the qname into localname and prefix*/
  SplitQName(qName, &prefix, &localName);

  res = (PyObject *)Document_CreateAttributeNS((PyDocumentObject *)self,
                                               namespaceUri,
                                               localName,
                                               prefix,
                                               NULL,
                                               &index);

  Py_DECREF(namespaceUri);
  Py_DECREF(prefix);
  Py_DECREF(localName);
  Py_DECREF(qName);

  return res;
}


static char PyDocument_createTextNode__doc__[] =
"Create a new text node";
PyObject *PyDocument_createTextNode(PyObject * self, PyObject * args) {

  PyObject *res;
  unsigned long index;
  PyObject *data;

  DOCUMENT_CHECK(self);

  if(!PyArg_ParseTuple(args, "O:createTextNode", &data))
    return NULL;

  data = PyUnicode_FromObject(data);
  if (!data) {
    return NULL;
  }

  res = (PyObject *)Document_CreateTextNode((PyDocumentObject *)self, data,
                                            &index);
  Py_DECREF(data);
  return res;
}


static char PyDocument_createProcessingInstruction__doc__[] =
"Create a new processing instruction node";
PyObject *PyDocument_createProcessingInstruction(PyObject * self, PyObject * args) {

  PyObject *res;
  unsigned long index;
  PyObject *target, *data;

  DOCUMENT_CHECK(self);

  if(!PyArg_ParseTuple(args,"OO:createTextNode",&target,&data))
    return NULL;

  
  target = PyUnicode_FromObject(target);
  if (!target) return NULL;


  data = PyUnicode_FromObject(data);
  if (!data) {
    Py_DECREF(target);
    return NULL;
  }

  res = (PyObject *)Document_CreateProcessingInstruction((PyDocumentObject *)self,
							 target, data, &index);
  Py_DECREF(data);
  Py_DECREF(target);
  return res;
}


static char PyDocument_createComment__doc__[] =
"Create a new comment node";
PyObject *PyDocument_createComment(PyObject * self, PyObject * args) {

  PyObject *res;
  unsigned long index;
  PyObject *data;

  if(!PyArg_ParseTuple(args,"O:createComment",&data))
    return NULL;

  data = PyUnicode_FromObject(data);
  if (!data) return NULL;

  res = (PyObject *)Document_CreateComment((PyDocumentObject *)self, data, 
					   &index);
  Py_DECREF(data);
  return res;
}


static char PyDocument_createDocumentFragment__doc__[] =
"Create a new document fragment node";
PyObject *PyDocument_createDocumentFragment(PyObject * self, PyObject * args) {

  PyObject *res;
  unsigned long index;

  DOCUMENT_CHECK(self);

  if(!PyArg_ParseTuple(args,":createComment"))
    return NULL;
  res = (PyObject *)Document_CreateDocumentFragment((PyDocumentObject *)self, &index);
  return res;
}


static char PyDocument_createNodeIterator__doc__[] =
"Create a new node iterator";
PyObject *PyDocument_createNodeIterator(PyObject * self, PyObject * args) {

  PyObject *res, *root, *whatToShow, *filter, *entityReferenceExpansion;
  PyObject *mod;

  DOCUMENT_CHECK(self);

  if(!PyArg_ParseTuple(args,"OOOO:createNodeIterator",&root, &whatToShow, &filter, &entityReferenceExpansion))
    return NULL;

  mod = PyImport_ImportModule("xml.dom.NodeIterator");
  if (mod) 
    res = PyObject_CallMethod(mod, "NodeIterator", "OOOO", root, whatToShow, filter, entityReferenceExpansion);
  else
    res = NULL;
  Py_XDECREF(mod);
  return res;
}


PyObject *PyDocument_importNode(PyObject *self, PyObject *args)
{
  int deep = 0;
  PyObject *node;

  DOCUMENT_CHECK(self);

  if(!PyArg_ParseTuple(args,"O|i:importNode", &node,&deep)) 
    return NULL;

  return (PyObject *)Node_CloneNode((PyNodeObject *)node,deep,(PyObject *)self);
}


static struct PyMethodDef Document_methods[] = { NODE_METHODS,
  {"importNode",   (PyCFunction)PyDocument_importNode,  METH_VARARGS, "Make a copy of a node from any document"}, \
  {"createElementNS",   (PyCFunction)PyDocument_createElementNS,  METH_VARARGS, PyDocument_createElementNS__doc__},
  {"createAttributeNS",   (PyCFunction)PyDocument_createAttributeNS,  METH_VARARGS, PyDocument_createAttributeNS__doc__},
  {"createTextNode",   (PyCFunction)PyDocument_createTextNode,  METH_VARARGS, PyDocument_createTextNode__doc__},
  {"createProcessingInstruction",   (PyCFunction)PyDocument_createProcessingInstruction,  METH_VARARGS, PyDocument_createProcessingInstruction__doc__},
  {"createComment",   (PyCFunction)PyDocument_createComment,  METH_VARARGS, PyDocument_createComment__doc__},
  {"createDocumentFragment",   (PyCFunction)PyDocument_createDocumentFragment,  METH_VARARGS, PyDocument_createDocumentFragment__doc__},
  {"createNodeIterator",   (PyCFunction)PyDocument_createNodeIterator,  METH_VARARGS, PyDocument_createNodeIterator__doc__},
  {NULL,     NULL}      /* sentinel */
};


/*Type methods*/

static PyObject *document_getattr(PyDocumentObject *self, char *name)
{
  PyObject *rt = NULL;
  int temp;

  DOCUMENT_CHECK(self);

  if (!strcmp(name,"ownerDocument")) {
    rt = Py_None;
  }
  else if (!strcmp(name,"documentElement")) {
    rt = self->documentElement;
  }
  else if (!strcmp(name,"implementation")) {
    rt = g_implementation;
  }
  else if (!strcmp(name,"doctype")) {
    rt = self->docType;
  }
  else if (!strcmp(name,"refUri")) {
    rt = (PyObject *)self->refUri;
  }
  else if (!strcmp(name,"nodeType")) {
    rt = g_documentNodeType;
  }
  else if (!strcmp(name,"nodeName")) {
    rt = g_documentNodeName;
  }
  else if (!strcmp(name,"childNodes")) {
    rt = self->childNodes;
  }
  else if (!strcmp(name,"lastChild")) {
    temp = PyList_GET_SIZE(self->childNodes);
    if (temp) {
      rt = PyList_GET_ITEM(self->childNodes,temp-1);
    } else {
      rt = Py_None;
    }
  }
  else if (!strcmp(name,"firstChild")) {
    if (PyList_GET_SIZE(self->childNodes)) {
      rt = PyList_GET_ITEM(self->childNodes,0);
    } else {
      rt = Py_None;
    }
  }
  if (rt) {
    Py_INCREF(rt);
    return rt;
  }
  return node_getattr((PyNodeObject*)self,name,Document_methods);
}

static int document_setattr(PyDocumentObject *self, char *name, PyObject *v)
{
  /* Set attribute 'name' to value 'v'. v==NULL means delete */
  if (v == NULL) {
    PyErr_Format(PyExc_AttributeError, 
                 "Cannot delete attribute '%.400s' on '%.50s' object",
                 name, self->ob_type->tp_name);
  } else {
    PyErr_Format(PyExc_AttributeError,
                 "Cannot set attribute '%.400s' on '%.50s' object", 
                 name, self->ob_type->tp_name);
  }
  return -1;
}

static void document_dealloc(PyDocumentObject *node)
{

  PyObject_GC_UnTrack((PyObject *) node);

  Py_XDECREF(node->childNodes);
  node->childNodes = NULL;

  Py_XDECREF(node->docType);
  node->docType = NULL;

  Py_XDECREF(node->documentElement);
  node->documentElement = NULL;

  Py_XDECREF(node->refUri);
  node->refUri = NULL;

  Py_XDECREF(node->internedUnicode);
  node->internedUnicode = NULL;

  node_dealloc((PyNodeObject*)node);
}


static int document_clear(PyDocumentObject *self)
{
  /*Called when we need to break cycles*/
  Py_XDECREF(self->childNodes);
  self->childNodes = NULL;
  Py_XDECREF(self->documentElement);
  self->documentElement = NULL;
  Py_XDECREF(self->docType);
  self->docType = NULL;
  return node_clear((PyNodeObject *)self);
}

static int document_traverse(PyDocumentObject *self, visitproc visit, void *arg)
{
  int rt;
  if (self->childNodes != NULL) {
    rt = visit(self->childNodes, arg);
    if (rt) return rt;
  }
  if (self->documentElement != NULL) {
    rt = visit(self->documentElement, arg);
    if (rt) return rt;
  }
  if (self->docType != NULL) {
    rt = visit(self->docType, arg);
    if (rt) return rt;
  }
  return node_traverse((PyNodeObject *)self,visit,arg);
}


static PyObject *document_repr(PyDocumentObject *document)
{
  char buf[256];
  sprintf(buf, "<cDocument at %p>", document);
  return PyString_FromString(buf);
}

static int document_print(PyDocumentObject *v, FILE *fp, int flags)
{
  PyObject *repr = document_repr(v);
  fprintf(fp, PyString_AS_STRING(repr));
  Py_DECREF(repr);
  return 0;
}

int document_test_ref_counts(PyObject *tester,PyDocumentObject *doc,int base) {
  /*To test a document, we must make sure that the ref counts to our self are correct
    To do this test (and count as we go) the number of children and attrs in the tree.  Also
    It is our job to test that the interned strings have the correct ref count */
  long childCount = 0;
  PyObject *internCtr = PyDict_New();
  int ctr;
  PyObject *curChild;
  int childBase;

#if 0 && !defined(DISABLE_INTERN)
  char buff[512];
  PyObject *v1, *v2, *temp;
  int i;
#endif

  for (ctr=0;ctr<PyList_GET_SIZE(doc->childNodes);ctr++) {
    curChild = PyList_GET_ITEM(doc->childNodes,ctr);
    if (curChild == doc->documentElement) {
      childBase = 1;
    } else {
      childBase = 0;
    }
    if (!node_test_ref_counts(tester,(PyNodeObject *)curChild,&childCount,internCtr,childBase)) {
      Py_DECREF(internCtr);
      return 0;
    };
  }

  PyObject_CallMethod(tester,"startTest","s","Node");
  /*Base additional*/
  if (!PyObject_CallMethod(tester,"compare","ll",childCount + base + PyList_GET_SIZE(doc->childNodes),doc->ob_refcnt)){
    Py_DECREF(internCtr);
    return 0;
  }
  PyObject_CallMethod(tester,"testDone","");

#if 0 && !defined(DISABLE_INTERN)
  PyObject_CallMethod(tester,"startTest","s","Intern Ref Count");
  i = 0;
  while (PyDict_Next(doc->internedUnicode, &i, &v1, &v2)) {
    /*Each entry in here will be 3 more then in the count.  2 for key,value in dict and one for key in the count dict*/
    temp = PyDict_GetItem(internCtr,v1);
    if (temp == NULL) {
      sprintf(buff,"String %s was in interned dict but not in the couting dict",PyString_AS_STRING(PyObject_Repr((PyObject *)v1)));
      PyObject_CallMethod(tester,"error","s",buff);
      return 0;
    }
    /*Make sure the internCtr + 3 = refCount of v1*/
    if (!PyObject_CallMethod(tester,"compare","lO",PyInt_AsLong(temp)+3,PyInt_FromLong((long)v1->ob_refcnt))) {
      return 0;
    }
  }

  printf("Check Backwards\n");

  PyObject_CallMethod(tester,"testDone","");
#endif

  Py_DECREF(internCtr);
  return 1;
}



PyTypeObject PyDomletteDocument_Type = {
    PyObject_HEAD_INIT(NULL)
    0,
    "cDomlette.Document",
    GC_TP_BASICSIZE(PyDocumentObject),
    0,
    (destructor)document_dealloc,   /*tp_dealloc*/
    (printfunc)document_print,  /*tp_print*/
    (getattrfunc)document_getattr,  /*tp_getattr*/
    (setattrfunc)document_setattr,  /*tp_setattr*/
    0,                          /*tp_compare*/
    (reprfunc)document_repr,    /*tp_repr*/
    0,                          /*tp_as_number*/
    0,              /*tp_as_sequence*/
    0,              /*tp_as_mapping*/
    0,                             /*tp_hash*/
    0,          /*tp_call*/
    0,          /*tp_str*/
    0,                          /* tp_getattro */
    0,                          /* tp_setattro */
    0,                          /* tp_as_buffer */
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,                          /* tp_doc */
    (traverseproc)document_traverse,  /* tp_traverse */
    (inquiry)document_clear,          /* tp_clear */
    0,                          /* tp_richcompare */
    0,                          /* tp_weaklistoffset */
};

